# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
# import os
# import sys
# sys.path.insert(0, os.path.abspath('.'))


# -- Project information -----------------------------------------------------
from kedro_pandera import __version__ as kp_version

project = "kedro-pandera"
copyright = "2023, Yolan Honoré-Rougé"
author = "Yolan Honoré-Rougé"

# The full version, including alpha/beta/rc tags
release = kp_version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_click",
    "sphinx_design",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.intersphinx",
    "sphinx.ext.extlinks",
    "sphinx.ext.viewcode",
    # "sphinx_autodoc_typehints",
    # "sphinx.ext.doctest",
    # "sphinx.ext.todo",
    # "sphinx.ext.coverage",
    # "sphinx.ext.mathjax",
    # "sphinx.ext.ifconfig",
    # "sphinx_sitemap",
    # "nbsphinx",
    "sphinx_copybutton",
    "sphinx_markdown_tables",
    "myst_parser",
    "sphinxawesome_theme",
    "sphinx_docsearch",  # before: "sphinxawesome_theme.docsearch",
    "sphinxawesome_theme.highlighting",
]

# enable autosummary plugin (table of contents for modules/classes/class
# methods)
autosummary_generate = True
autosummary_generate_overwrite = False
napoleon_include_init_with_doc = True

# enable documentation in markdown
source_suffix = {".rst": "restructuredtext", ".md": "markdown"}


# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build"]


# -- Options for HTML output -------------------------------------------------
# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinxawesome_theme"

# useful to create dropdown with the name of the directory as the section name
# see https://stackoverflow.com/questions/36925871/toctree-nested-drop-down:
html_theme_options = {"collapse_navigation": False}


myst_heading_anchors = 5



import logging
from typing import Any, Dict, Set

from kedro.framework.context import KedroContext
from kedro.framework.hooks import hook_impl
from kedro.io import DataCatalog
from kedro.pipeline.node import Node
from pandera.errors import SchemaError

from kedro_pandera.framework.config.resolvers import (
    resolve_dataframe_model,
    resolve_interpolated_yaml_schema,
    resolve_yaml_schema,
)

# if we do not import ``frictionless`` manually here, we get
# >  ImportError: ('# ERROR: failed to register fsspec file#systems', TypeError("argument of type '_Cached' is not iterable"))
# if we try to use ``Schema.to_yaml()`` after ``context.catalog``
# The command ``kedro pandera infer-schema -d example_iris_data``
# then raises the very useless error:
# > ImportError: IO and formatting requires 'pyyaml', 'black' and 'frictionless'to be installed.
# > You can install pandera together with the IO dependencies with:
# > pip install pandera[io]
# despite all the dependencies being properly installed


class PanderaHook:
    def __init__(self) -> None:
        self._validated_datasets: Set[str] = set()

    @property
    def _logger(self) -> logging.Logger:
        return logging.getLogger(__name__)

    @hook_impl
    def after_context_created(
        self,
        context: KedroContext,
    ) -> None:
        """Hooks to be invoked after a `KedroContext` is created. This is the earliest
        hook triggered within a Kedro run. The `KedroContext` stores useful information
        such as `credentials`, `config_loader` and `env`.
        Args:
            context: The context that was created.
        """
        context.config_loader._register_new_resolvers(
            {
                "pa.dict": resolve_yaml_schema,
                "pa.yaml": resolve_interpolated_yaml_schema,
                "pa.python": resolve_dataframe_model,
            }
        )

    def _validate_datasets(
        self, node: Node, catalog: DataCatalog, datasets: Dict[str, Any]
    ):
        for name, data in datasets.items():
            dataset = catalog._datasets.get(name)
            metadata = getattr(dataset, "metadata", None)
            if (
                metadata is not None
                and "pandera" in metadata
                and name not in self._validated_datasets
            ):
                try:
                    metadata["pandera"]["schema"].validate(data)
                    self._validated_datasets.add(name)
                except SchemaError as err:
                    self._logger.error(
                        f"Dataset '{name}' pandera validation failed before running '{node.name}', see details in the error message. "
                    )
                    raise err
                except Exception as err:
                    self._logger.error(
                        f"Dataset '{name}' validation raised an unknown error before running '{node.name}'"
                    )
                    raise err

                self._logger.info(
                    f"(kedro-pandera) Dataset '{name}' was successfully validated with pandera"
                )

    @hook_impl
    def before_node_run(  # noqa : PLR0913
        self,
        node: Node,
        catalog: DataCatalog,
        inputs: Dict[str, Any],
        is_async,
        session_id,
    ):
        self._validate_datasets(node, catalog, inputs)

    @hook_impl
    def after_node_run(self, node: Node, catalog: DataCatalog, outputs: Dict[str, Any]):
        self._validate_datasets(node, catalog, outputs)


pandera_hook = PanderaHook()



import pytest
from kedro.framework.hooks import _create_hook_manager
from kedro.framework.hooks.manager import _register_hooks
from kedro.io import DataCatalog, LambdaDataset
from kedro.pipeline import node, pipeline
from kedro.runner import SequentialRunner
from kedro_datasets.pandas import CSVDataset
from pandera.errors import SchemaError
from pandera.io import from_yaml

from kedro_pandera.framework.hooks.pandera_hook import PanderaHook


def _get_test_catalog(csv_file, schema_file):
    test_schema = from_yaml(schema_file)
    test_catalog = DataCatalog(
        {
            "iris": CSVDataset(
                filepath=csv_file,
                metadata={"pandera": {"schema": test_schema}},
            ),
        }
    )
    return test_catalog


def _get_test_hook():
    test_hook = PanderaHook()
    return test_hook


def _run_hook(csv_file, schema_file):
    test_catalog = _get_test_catalog(csv_file, schema_file)
    test_hook = _get_test_hook()
    test_inputs = {"iris": test_catalog.load("iris")}
    test_node = node(
        name="test_node", func=lambda iris: True, inputs=["iris"], outputs=None
    )
    converted_inputs = test_hook.before_node_run(
        node=test_node,
        catalog=test_catalog,
        inputs=test_inputs,
        is_async=False,
        session_id=0,
    )
    return test_inputs, converted_inputs


def test_hook():
    _run_hook(
        csv_file="tests/data/iris.csv",
        schema_file="tests/data/iris_schema.yml",
    )


def test_hook_validation_error():
    with pytest.raises(SchemaError):
        _run_hook(
            csv_file="tests/data/iris.csv",
            schema_file="tests/data/iris_schema_fail.yml",
        )


def test_hook_unexpected_error():
    test_catalog = _get_test_catalog(
        csv_file="tests/data/iris.csv", schema_file="tests/data/iris_schema.yml"
    )
    test_hook = _get_test_hook()
    test_inputs = {"iris": []}
    test_node = node(
        name="test_node", func=lambda iris: True, inputs=["iris"], outputs=None
    )
    with pytest.raises(Exception):
        test_hook.before_node_run(
            node=test_node,
            catalog=test_catalog,
            inputs=test_inputs,
            is_async=False,
            session_id=0,
        )


def test_hook_output_validation():
    test_catalog = _get_test_catalog(
        csv_file="tests/data/iris.csv", schema_file="tests/data/iris_schema.yml"
    )
    test_hook = _get_test_hook()
    test_outputs = {"iris": test_catalog.load("iris")}
    test_node = node(
        name="test_node", func=lambda iris: True, inputs=["iris"], outputs=None
    )
    test_hook.after_node_run(
        node=test_node,
        catalog=test_catalog,
        outputs=test_outputs,
    )


def test_validate_only_once(caplog):
    test_catalog = _get_test_catalog(
        csv_file="tests/data/iris.csv", schema_file="tests/data/iris_schema.yml"
    )
    test_hook = _get_test_hook()
    test_inputs = {"iris": test_catalog.load("iris")}
    test_node = node(
        name="test_node", func=lambda iris: True, inputs=["iris"], outputs=None
    )
    test_hook.before_node_run(
        node=test_node,
        catalog=test_catalog,
        inputs=test_inputs,
        is_async=False,
        session_id=0,
    )
    test_hook.before_node_run(
        node=test_node,
        catalog=test_catalog,
        inputs=test_inputs,
        is_async=False,
        session_id=0,
    )
    # should only be validated once
    assert caplog.text.count("successfully validated") == 1


def test_no_exception_on_memory_dataset_output():
    test_hook_manager = _create_hook_manager()
    test_hook = _get_test_hook()
    HOOKS = (test_hook,)
    _register_hooks(test_hook_manager, HOOKS)
    test_catalog = DataCatalog(
        {
            "Input": LambdaDataset(load=lambda: "data", save=lambda data: None),
            "Output": LambdaDataset(load=lambda: "data", save=lambda data: None),
        }
    )
    test_pipeline = pipeline(
        [
            node(func=lambda x: x, inputs="Input", outputs="MemOutput", name="node1"),
            node(func=lambda x: x, inputs="MemOutput", outputs="Output", name="node2"),
        ]
    )
    assert test_hook_manager.is_registered(test_hook)
    SequentialRunner().run(test_pipeline, test_catalog, hook_manager=test_hook_manager)



from pathlib import Path

import pytest
from cookiecutter.main import cookiecutter
from kedro import __version__ as kedro_version
from kedro.framework.cli.starters import TEMPLATE_PATH

# cloned with version 0.18.12
# modified to have configloader by default
# removed tests to avoid pytest conflicts
PANDAS_IRIS_PATH = (Path(__file__).parent / "_data" / "pandas-iris").as_posix()

_FAKE_PROJECT_NAME = "fake project"
_FAKE_REPO_NAME = "fake-project"
_FAKE_PACKAGE_NAME = "fake_project"


@pytest.fixture
def kedro_project(tmp_path):
    # TODO : this is also an integration test since this depends from the kedro version
    config = {
        "output_dir": tmp_path,
        "kedro_version": kedro_version,
        "repo_name": _FAKE_REPO_NAME,
        "project_name": _FAKE_PROJECT_NAME,
        "python_package": _FAKE_PACKAGE_NAME,
    }

    cookiecutter(
        str(TEMPLATE_PATH),
        output_dir=config["output_dir"],
        no_input=True,
        extra_context=config,
        accept_hooks=False,
    )
    return tmp_path / _FAKE_REPO_NAME


@pytest.fixture
def kedro_project_iris(tmp_path):
    # TODO : this is also an integration test since this depends from the kedro version

    config = {
        "output_dir": tmp_path,
        "kedro_version": kedro_version,
        "repo_name": _FAKE_REPO_NAME,
        "project_name": _FAKE_PROJECT_NAME,
        "python_package": _FAKE_PACKAGE_NAME,
    }
    cookiecutter(
        template=PANDAS_IRIS_PATH,
        output_dir=config["output_dir"],
        no_input=True,
        extra_context=config,
    )

    return tmp_path / _FAKE_REPO_NAME



import sys
from contextlib import contextmanager

import pytest
import yaml
from omegaconf import OmegaConf
from pandera import DataFrameSchema
from pandera.errors import SchemaDefinitionError

from kedro_pandera.framework.config.resolvers import (
    resolve_dataframe_model,
    resolve_interpolated_yaml_schema,
    resolve_yaml_schema,
)


@contextmanager
def register_temporary_resolver(name, resolver):
    _flag_restore = True if OmegaConf.has_resolver(name) else False
    if _flag_restore:
        old_resolver = OmegaConf._get_resolver(name)

    OmegaConf.register_new_resolver(name, resolver, replace=True)

    yield

    if _flag_restore:
        OmegaConf.register_new_resolver(name, old_resolver, replace=True)
    else:
        OmegaConf.clear_resolver(name)


MINIMAL_SCHEMA_EXAMPLE = yaml.safe_load(
    """
schema_type: dataframe
version: 0.16.1
columns:
sepal_length:
    title: null
    description: null
    dtype: float64
    nullable: false
    checks:
    greater_than_or_equal_to: 4.3
    less_than_or_equal_to: 7.9
    unique: false
    coerce: false
    required: true
    regex: false
checks: null
index:
- title: null
description: null
dtype: int64
nullable: false
checks:
    greater_than_or_equal_to: 0.0
    less_than_or_equal_to: 149.0
name: null
unique: false
coerce: false
dtype: null
coerce: true
strict: false
name: null
ordered: false
unique: null
report_duplicates: all
unique_column_names: false
add_missing_columns: false
title: null
description: null
"""
)


def test_resolve_yaml_schema():
    df_schema = resolve_yaml_schema(MINIMAL_SCHEMA_EXAMPLE)
    assert isinstance(df_schema, DataFrameSchema)


def test_resolve_yaml_schema_fails_with_invalid_yaml():
    with pytest.raises(SchemaDefinitionError):
        _ = resolve_yaml_schema("foobar")


def test_resolve_yaml_schema_works_as_resolver():
    # maybe not a useful test: after all, this is OmegaConf responsibility
    with register_temporary_resolver("pa.dict", resolve_yaml_schema):
        config = OmegaConf.create(
            {
                "my_data": {
                    "type": "pandas.CSVDataset",
                    "filepath": "path/to/data.csv",
                    "metadata": {
                        "pandera": {
                            "schema": "${pa.dict:${oc.select:_data_schema,null}}",
                        }
                    },
                },
                "_data_schema": MINIMAL_SCHEMA_EXAMPLE,
            }
        )
        assert isinstance(config.my_data.metadata.pandera.schema, DataFrameSchema)


def test_resolve_interpolated_yaml_schema():
    config = OmegaConf.create(
        {
            "my_data": {
                "type": "pandas.CSVDataset",
                "filepath": "path/to/data.csv",
                "metadata": {
                    "pandera": {
                        "schema": "${_data_schema}",
                    }
                },
            },
            "_data_schema": MINIMAL_SCHEMA_EXAMPLE,
        }
    )
    df_schema = resolve_interpolated_yaml_schema("_data_schema", _parent_=config)
    assert isinstance(df_schema, DataFrameSchema)


@pytest.fixture
def mock_dataframe_model(mocker):
    sys.modules["dummy"] = mocker.Mock()  # This is the easiest way to mock a module
    sys.modules["dummy.schema"] = mocker.Mock()
    sys.modules["dummy.schema"].dataframe_model = mocker.Mock()
    yield sys.modules["dummy.schema"].dataframe_model
    del sys.modules["dummy"]
    del sys.modules["dummy.schema"]


def test_resolve_dataframe_model(mock_dataframe_model):
    model = resolve_dataframe_model("dummy.schema.dataframe_model")
    assert model == mock_dataframe_model



import pytest
from click.testing import CliRunner

from kedro_pandera.framework.cli.cli import infer_dataset_schema


@pytest.mark.parametrize("flag_extension", ["--python", "--yaml"])
def test_schema_is_persisted_with_correct_extension(
    monkeypatch, kedro_project_iris, flag_extension
):
    monkeypatch.chdir(kedro_project_iris)
    # launch the command to initialize the project
    cli_runner = CliRunner()
    result = cli_runner.invoke(
        infer_dataset_schema,
        ["-d", "example_iris_data", flag_extension],
    )
    file_extension = "py" if flag_extension == "--python" else "yml"

    assert result.exit_code == 0
    assert (
        kedro_project_iris
        / "conf"
        / "base"
        / f"catalog_example_iris_data.{file_extension}"
    ).is_file()



import re

from click.testing import CliRunner

from kedro_pandera.framework.cli.cli import pandera_commands as cli_pandera


def extract_cmd_from_help(msg):
    # [\s\S] is used instead of "." to match any character including new lines
    cmd_txt = re.search((r"(?<=Commands:)([\s\S]+)$"), msg).group(1)
    cmd_list_detailed = cmd_txt.split("\n")

    cmd_list = []
    for cmd_detailed in cmd_list_detailed:
        cmd_match = re.search(r"[\w-]+(?=  )", string=cmd_detailed)
        if cmd_match is not None:
            cmd_list.append(cmd_match.group(0))
    return cmd_list


def test_pandera_commands_inside_kedro_project(monkeypatch, kedro_project):
    monkeypatch.chdir(kedro_project)
    # launch the command to initialize the project
    cli_runner = CliRunner()
    result = cli_runner.invoke(cli_pandera)
    assert {"infer"} == set(extract_cmd_from_help(result.output))



from __future__ import annotations

from pathlib import Path

import click
import frictionless  # noqa: F401
import yaml
from kedro.framework.project import settings
from kedro.framework.session import KedroSession
from kedro.framework.startup import bootstrap_project
from pandera import infer_schema
from pandera.io import serialize_schema

# if we do not import ``frictionless`` manually here, we get
# >  ImportError: ('# ERROR: failed to register fsspec file#systems', TypeError("argument of type '_Cached' is not iterable"))
# if we try to use ``Schema.to_yaml()`` after ``context.catalog``
# The command ``kedro pandera infer-schema -d example_iris_data``
# then raises the very useless error:
# > ImportError: IO and formatting requires 'pyyaml', 'black' and 'frictionless'to be installed.
# > You can install pandera together with the IO dependencies with:
# > pip install pandera[io]
# despite all the dependencies being properly installed


@click.group(name="pandera")
def commands():
    """Kedro plugin for interactions with pandera"""
    pass  # pragma: no cover


@commands.group(name="pandera")
def pandera_commands():
    """Use pandera-specific commands inside kedro project."""
    pass  # pragma: no cover


@pandera_commands.command(name="infer")
@click.option(
    "--dataset",
    "-d",
    "dataset_name",
    help="The name of the dataset whom schema will be inferred.",
)
@click.option(
    "--python/--yaml",
    "use_python_ext",
    is_flag=True,
    default=False,
    help="Should the schema be written to a 'yaml' or a 'python' file? ",
)
@click.option(
    "--env",
    "-e",
    default="base",
    help="The kedro environment where the dataset to retrieve is available. Default to 'local'",
)
@click.option(
    "--outfile",
    "-o",
    default=None,
    help="The name of the file where the schema will be stored. Its extension must be '.yml' or '.py'. Default to 'dataset_name.yml'. ",
)
def infer_dataset_schema(
    dataset_name: str,
    use_python_ext: bool = False,
    env: str = "base",
    outfile: str | None = None,
):
    """Infer the schema of a dataset and dump it in a catalog file
    so that it will enable validation at runtime.
    """

    project_path = Path().cwd()
    bootstrap_project(project_path)
    with KedroSession.create(
        project_path=project_path,
        env=env,
    ) as session:
        context = session.load_context()
        catalog = context.catalog
        data = catalog.load(dataset_name)
        data_schema = infer_schema(data)
        file_extension = "py" if use_python_ext else "yml"
        filename = f"catalog_{dataset_name}.{file_extension}"
        # TODO: decide if we should store in env or base?
        filepath = (project_path / settings.CONF_SOURCE / env / filename).as_posix()
        if use_python_ext:
            data_schema.to_script(filepath)
        else:
            statistics = serialize_schema(data_schema)
            with open(filepath, "w", encoding="utf-8") as fhandler:
                statistics = serialize_schema(data_schema)
                yaml.safe_dump(
                    {f"_{dataset_name}_schema": statistics},
                    stream=fhandler,
                    sort_keys=False,
                )
                # data_schema.to_yaml(filepath)

        click.secho(f"The schema is written to {filepath}", fg="green")

        # TODO: should we add a metadata key to the catalog file?
        # I hate the idea of modifying a config flag on the fly,
        # but it will be very confusing for users to have to add it manually each time, and they will likely forget
        # This suggestion :https://github.com/kedro-org/kedro/issues/1076#issuecomment-1490277770
        # seems ultimately the best solution, but we need the omegaconfigloader to be more stable
        # https://github.com/kedro-org/kedro/issues/1076#issuecomment-1491008769

        # TODO: where should we store the output file? in a config folder?
        # What will happen in package mode when running with "--conf-source=path/to/conf"?

        # TODO: should we check if the file already exists before overwriting ?
        # TODO: should we compare an existing file with the inferred one to spot the differences?

        # TODO: Define if we should have a flag --py //--yml or infer from the file extension



from omegaconf import Container
from omegaconf.resolvers.oc import select
from pandera import DataFrameSchema
from pandera.io import deserialize_schema


# will be registered as pa.dict
def resolve_yaml_schema(schema: str) -> DataFrameSchema:
    # intended use in catalog is :
    # schema: ${pa.dict:${oc.select:_example_iris_data_schema,null}}

    # This will fail (e.g. with the CLI "kedro pandera infer") if the key does not exist yet, that's why we need a default "null"
    pandera_schema = deserialize_schema(schema)
    return pandera_schema


# will be registered as pa.yaml
def resolve_interpolated_yaml_schema(
    key: str, *, _parent_: Container
) -> DataFrameSchema:
    # the goal is to simplify pa.dict to avoid a complex combination of resolvers
    #  base way with resolve_yaml_schema : ${pa.dict:${oc.select:_example_iris_data_schema,null}}
    # simplified way with resolve_interpolated_yaml_schema : ${pa.yaml:_example_iris_data_schema}

    schema = select(key, default=None, _parent_=_parent_)
    pandera_schema = resolve_yaml_schema(schema)
    return pandera_schema


def resolve_dataframe_model(schema_name):
    import importlib

    module, _, schema = schema_name.rpartition(".")
    module = importlib.import_module(module)
    return getattr(module, schema)



import pathlib

from setuptools import find_packages, setup

NAME = "kedro_pandera"
HERE = pathlib.Path(__file__).parent


def _parse_requirements(path, encoding="utf-8"):
    with open(path, encoding=encoding) as file_handler:
        requirements = [
            x.strip() for x in file_handler if x.strip() and not x.startswith("-r")
        ]
    return requirements


# get the dependencies and installs
base_requirements = _parse_requirements((HERE / "requirements.txt").as_posix())


# Get the long description from the README file
with open((HERE / "README.md").as_posix(), encoding="utf-8") as file_handler:
    README = file_handler.read()


setup(
    name=NAME,
    version="0.2.2",  # this will be bumped automatically by bump2version
    description="A kedro plugin to use pandera in your kedro projects",
    license="Apache Software License (Apache 2.0)",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Galileo-Galilei/kedro-pandera",
    python_requires=">=3.8, <3.12",
    packages=find_packages(exclude=["docs*", "tests*"]),
    setup_requires=["setuptools_scm"],
    include_package_data=True,
    install_requires=base_requirements,
    extras_require={
        "doc": [
            "sphinx>=4.5.0,<7.3.0",  # https://github.com/kai687/sphinxawesome-theme/issues/1464
            "sphinxawesome-theme",
            "sphinx-docsearch",
            "sphinx-markdown-tables~=0.0.15",
            "sphinx-click>=3.1,<6.0",
            "sphinx_copybutton~=0.5.0",
            # "sphinx-sitemap",
            "sphinx-design",
            "myst-parser>=0.17.2,<2.1.0",
        ],
        "test": [
            "ruff>=0.4.0, <0.5.0",
            "pytest>=7.0.0, <8.0.0",
            "pytest-cov>=4.0.0, <5.0.0",
            "pytest-mock",
            "pre-commit>=2.0.0,<4.0.0",
            "cookiecutter",
            "kedro-datasets",
        ],
        "dev": [
            "pre-commit>=2.0.0,<4.0.0",
            "jupyter>=1.0.0,<2.0.0",
        ],
    },
    author="Yolan Honoré-Rougé",
    entry_points={
        "kedro.project_commands": [
            "kedro_pandera =  kedro_pandera.framework.cli.cli:commands"
        ],
        "kedro.hooks": [
            "pandera_hook = kedro_pandera.framework.hooks.pandera_hook:pandera_hook",
        ],
    },
    zip_safe=False,
    keywords="",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Framework :: Kedro",
        "Environment :: Plugins",
        "Intended Audience :: Developers",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS",
        "Operating System :: POSIX :: Linux",
    ],
)



"""
This is a boilerplate pipeline
generated using Kedro {{ cookiecutter.kedro_version }}
"""

from kedro.pipeline import Pipeline, node, pipeline

from .nodes import make_predictions, report_accuracy, split_data


def create_pipeline(**kwargs) -> Pipeline:
    return pipeline(
        [
            node(
                func=split_data,
                inputs=["example_iris_data", "parameters"],
                outputs=["X_train", "X_test", "y_train", "y_test"],
                name="split",
            ),
            node(
                func=make_predictions,
                inputs=["X_train", "X_test", "y_train"],
                outputs="y_pred",
                name="make_predictions",
            ),
            node(
                func=report_accuracy,
                inputs=["y_pred", "y_test"],
                outputs=None,
                name="report_accuracy",
            ),
        ]
    )



"""
This is a boilerplate pipeline
generated using Kedro {{ cookiecutter.kedro_version }}
"""

import logging
from typing import Any

import numpy as np
import pandas as pd


def split_data(
    data: pd.DataFrame, parameters: dict[str, Any]
) -> tuple[pd.DataFrame, pd.DataFrame, pd.Series, pd.Series]:
    """Splits data into features and target training and test sets.

    Args:
        data: Data containing features and target.
        parameters: Parameters defined in parameters.yml.
    Returns:
        Split data.
    """

    data_train = data.sample(
        frac=parameters["train_fraction"], random_state=parameters["random_state"]
    )
    data_test = data.drop(data_train.index)

    X_train = data_train.drop(columns=parameters["target_column"])
    X_test = data_test.drop(columns=parameters["target_column"])
    y_train = data_train[parameters["target_column"]]
    y_test = data_test[parameters["target_column"]]

    return X_train, X_test, y_train, y_test


def make_predictions(
    X_train: pd.DataFrame, X_test: pd.DataFrame, y_train: pd.Series
) -> pd.Series:
    """Uses 1-nearest neighbour classifier to create predictions.

    Args:
        X_train: Training data of features.
        y_train: Training data for target.
        X_test: Test data for features.

    Returns:
        y_pred: Prediction of the target variable.
    """

    X_train_numpy = X_train.to_numpy()
    X_test_numpy = X_test.to_numpy()

    squared_distances = np.sum(
        (X_train_numpy[:, None, :] - X_test_numpy[None, :, :]) ** 2, axis=-1
    )
    nearest_neighbour = squared_distances.argmin(axis=0)
    y_pred = y_train.iloc[nearest_neighbour]
    y_pred.index = X_test.index

    return y_pred


def report_accuracy(y_pred: pd.Series, y_test: pd.Series):
    """Calculates and logs the accuracy.

    Args:
        y_pred: Predicted target.
        y_test: True target.
    """
    accuracy = (y_pred == y_test).sum() / len(y_test)
    logger = logging.getLogger(__name__)
    logger.info("Model has accuracy of %.3f on test data.", accuracy)
